/*
 * Aiptek PocketDV 3300 module version 0.2
 * released under GPL
 *
 * Allows you to use your webcam with V4L applications
 * This is still in heavy developpement !
 *
 * http://royale.zerezo.com/dv3300/
 *
 * Heavily inspired by "usb-skeleton.c", "vicam.c" and "spca50x.c" drivers
 * It also uses the JPEG decoder of the "bootsplash" project (a bit hacked)
 */


#include <linux/kernel.h>
#include <linux/wrapper.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/videodev.h>
#include <linux/usb.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>

#include "decode-jpg.h"
#include "decode-jpg.c"
#include "osd.c"


/* USB ids */
#define USB_DV3300_VENDOR_ID  0x08ca
#define USB_DV3300_PRODUCT_ID 0x0109


/* Version Information */
#define DRIVER_VERSION "v0.2"
#define DRIVER_AUTHOR "Antoine Jacquet, http://royale.zerezo.com/"
#define DRIVER_DESC "Aiptek PocketDV 3300"


/* Camera */
#define VID_HARDWARE_DV3300 0xff
#define DV3300_FRAMES 1
#define DV3300_MAX_FRAME_SIZE 640*480*3

#define BUFFER_SIZE 0x1000
#define CTRL_TIMEOUT 500


/* debug macro */
#define DBG(x...) if (debug) info(x)


/* Module parameters */
static int debug=0;
static int osd=0;


/* Module parameters interface */
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug level");
MODULE_PARM(osd, "i");
MODULE_PARM_DESC(osd, "OSD frame count");


/* Camera stuff */
struct dv3300
{
  struct usb_device *udev;			/* save off the usb device pointer */
  struct usb_interface *interface;		/* the interface for this device */
	struct video_device vdev;	// v4l video device
  u8 *framebuf;
  int nb;
  unsigned char *buffer;
  unsigned char *jpeg;
  struct jpeg_decdata decdata;
  int skip;
  int brightness;
  int contrast;
};


static int
send_control_msg(struct usb_device *udev, u8 request, u16 value, u16 index,
		 unsigned char *cp, u16 size)
{
	int status;

	// for reasons not yet known to me, you can't send USB control messages
	// with data in the module (if you are compiled as a module).  Whatever
	// the reason, copying it to memory allocated as kernel memory then
	// doing the usb control message fixes the problem.

	unsigned char *transfer_buffer = kmalloc(size, GFP_KERNEL);
	memcpy(transfer_buffer, cp, size);

	status = usb_control_msg(udev,
				 usb_sndctrlpipe(udev, 0),
				 request,
				 USB_DIR_OUT | USB_TYPE_VENDOR |
				 USB_RECIP_DEVICE, value, index,
				 transfer_buffer, size, CTRL_TIMEOUT);

	kfree(transfer_buffer);

	if (status < 0) {
		info("Failed sending control message, error %d.\n",status);
	}

	return status;
}


/* Control messages sent to the camera to initialize it and launch the capture */

static unsigned char init1[]={0,0,0,0,0,0,0,0,0,0};
static unsigned char init2[]={0,0,0,0,0,0};
static unsigned char init3[]={0,0};

static struct {
  int value;
  int size;
  char *bytes;
} init[]={
      {0x1f30, 0, NULL},
      {0xd000, 0, NULL},
      {0x3370, sizeof(init1), init1},
      {0x2000, 0, NULL},
      {0x2f0f, 0, NULL},
      {0x2610, sizeof(init2), init2},
      {0xe107, 0, NULL},
      {0x2502, 0, NULL},
      {0x1f70, 0, NULL},
      {0xd000, 0, NULL},
      {0x9a01, sizeof(init3), init3},
      {-1,-1,NULL}
};


/* JPEG fixed data in header (Huffman table, etc) */

static unsigned char header1[]={
                          0xFF,0xD8, /* 0xFF,0xE0,0x00,0x10,'J','F','I','F',0x00,0x01,0x01,0x00,0x33,0x8A,0x00,0x00,0x33,0x88, */
                          0xFF,0xDB,0x00,0x84
                        };
static unsigned char header2[]={
                          0xFF,0xC4,0x00,0x1F,0x00,0x00,0x01,0x05,0x01,0x01,0x01,0x01,0x01,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0xFF,0xC4,0x00,0xB5,0x10,0x00,0x02,0x01,0x03,0x03,0x02,0x04,0x03,0x05,0x05,0x04,0x04,0x00,0x00,0x01,0x7D,0x01,0x02,0x03,0x00,0x04,0x11,0x05,0x12,0x21,0x31,0x41,0x06,0x13,0x51,0x61,0x07,0x22,0x71,0x14,0x32,0x81,0x91,0xA1,0x08,0x23,0x42,0xB1,0xC1,0x15,0x52,0xD1,0xF0,0x24,0x33,0x62,0x72,0x82,0x09,0x0A,0x16,0x17,0x18,0x19,0x1A,0x25,0x26,0x27,0x28,0x29,0x2A,0x34,0x35,0x36,0x37,0x38,0x39,0x3A,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4A,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5A,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6A,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFF,0xC4,0x00,0x1F,0x01,0x00,0x03,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0xFF,0xC4,0x00,0xB5,0x11,0x00,0x02,0x01,0x02,0x04,0x04,0x03,0x04,0x07,0x05,0x04,0x04,0x00,0x01,0x02,0x77,0x00,0x01,0x02,0x03,0x11,0x04,0x05,0x21,0x31,0x06,0x12,0x41,0x51,0x07,0x61,0x71,0x13,0x22,0x32,0x81,0x08,0x14,0x42,0x91,0xA1,0xB1,0xC1,0x09,0x23,0x33,0x52,0xF0,0x15,0x62,0x72,0xD1,0x0A,0x16,0x24,0x34,0xE1,0x25,0xF1,0x17,0x18,0x19,0x1A,0x26,0x27,0x28,0x29,0x2A,0x35,0x36,0x37,0x38,0x39,0x3A,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4A,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5A,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6A,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,
                          0xFF,0xC0,0x00,0x11,0x08,0x00,0xF0,0x01,0x40,0x03,0x01,0x21,0x00,0x02,0x11,0x01,0x03,0x11,0x01,
                          0xFF,0xDA,0x00,0x0C,0x03,0x01,0x00,0x02,0x11,0x03,0x11,0x00,0x3F,0x00
                        };
static unsigned char header3;



static int dv3300_open(struct video_device *dev,int flags)
{
	struct dv3300 *cam = (struct dv3300 *) dev->priv;
  struct usb_device *udev = cam->udev;
  int i,n;

  DBG("dv3300_open\n");

  for (i=0;init[i].size!=-1;i++)
  {
    n=send_control_msg(udev, /*0x40,*/ 1, init[i].value, 0, init[i].bytes, init[i].size);
    if (n<0)
    {
      info("error during initialisation");
      return n;
    }
  }

  return 0;
}

static void dv3300_close(struct video_device *dev)
{
  DBG("dv3300_close\n");
}

static void read_frame(struct dv3300 *cam, int framenum);

static long dv3300_read(struct video_device *dev,char *buf,unsigned long count,int noblock)
{
	struct dv3300 *cam = dev->priv;
	DBG("dv3300_read: read %d bytes.\n", (int) count);

	if (!buf)
		return -EINVAL;

	if (!count)
		return -EINVAL;

  read_frame(cam,0);
  
  copy_to_user(buf,cam->framebuf,count);

  return count;
}

static int dv3300_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	// TODO: allocate the raw frame buffer if necessary
	unsigned long start = (unsigned long) adr;
	unsigned long page, pos;
	struct dv3300 *cam = dev->priv;

	if (!cam)
		return -ENODEV;

	DBG("dv3300_mmap: %ld\n", size);

	/* We let mmap allocate as much as it wants because Linux was adding 2048 bytes
	 * to the size the application requested for mmap and it was screwing apps up.
	 if (size > VICAM_FRAMES*VICAM_MAX_FRAME_SIZE)
	 return -EINVAL;
	 */

	/* make this _really_ smp-safe */
	/*if (down_interruptible(&cam->busy_lock))
		return -EINTR;*/

	if (!cam->framebuf) {	/* we do lazy allocation */
		cam->framebuf =
		    (u8*)usbvideo_rvmalloc(DV3300_MAX_FRAME_SIZE * DV3300_FRAMES);
		if (!cam->framebuf) {
			//up(&cam->busy_lock);
			return -ENOMEM;
		}
	}

	pos = (unsigned long) (cam->framebuf);
	while (size > 0) {
		page = usbvideo_kvirt_to_pa(pos);
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED)) {
			//up(&cam->busy_lock);
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	//up(&cam->busy_lock);

	return 0;
}

static int dv3300_video_init(struct video_device *vdev)
{
  DBG("dv3300_video_init\n");
  return 0;
}

static long dv3300_write(struct video_device *v,const char *buf,unsigned long count,int nonblock)
{
  DBG("dv3300_write\n");
	return -EINVAL;
}

static void read_frame(struct dv3300 *cam, int framenum)
{
	int i,n,temp;
  int head=0;
	int actual_length;
  unsigned char *ptr=NULL;
  unsigned char swap;
  char text[20]="FRAME : 000000";
  unsigned char *pos;

  do
  {
    n=usb_bulk_msg(cam->udev, usb_rcvbulkpipe(cam->udev, 0x81), cam->buffer, BUFFER_SIZE, &actual_length, CTRL_TIMEOUT);
    DBG("buffer : %d %d",cam->buffer[0],cam->buffer[1]);
    DBG("bulk : n=%d size=%d",n,actual_length);
    if (n<0) 
    {
      info("error reading bulk msg");
      return;
    }
    if (actual_length<0 || actual_length>BUFFER_SIZE)
    {
      info("wrong number of bytes");
      return;
    }
    
    /* swap to good indian */
    for (i=0;i<BUFFER_SIZE;i+=2)
    {
      swap=cam->buffer[i];
      cam->buffer[i]=cam->buffer[i+1];
      cam->buffer[i+1]=swap;
    }
    
    /* write the JPEG header */
    if (!head)
    {
      DBG("jpeg header");
      ptr=cam->jpeg;
      memcpy(ptr,header1,sizeof(header1));
      ptr+=sizeof(header1);
      header3=0;
      memcpy(ptr,&header3,1);
      ptr++;
      memcpy(ptr,cam->buffer,64);
      ptr+=64;
      header3=1;
      memcpy(ptr,&header3,1);
      ptr++;
      memcpy(ptr,cam->buffer+64,64);
      ptr+=64;
      memcpy(ptr,header2,sizeof(header2));
      ptr+=sizeof(header2);
      memcpy(ptr,cam->buffer+128,actual_length-128);
      ptr+=actual_length-128;
      head=1;
      
      DBG("header : %d %d %d %d %d %d %d %d %d",cam->buffer[0],cam->buffer[1],cam->buffer[2],cam->buffer[3],cam->buffer[4],cam->buffer[5],cam->buffer[6],cam->buffer[7],cam->buffer[8]);
    }
    else
    {
      memcpy(ptr,cam->buffer,actual_length);
      ptr+=actual_length;
    }
  }
  while (actual_length==BUFFER_SIZE);

  DBG("jpeg : %d %d %d %d %d %d %d %d %d",cam->jpeg[0],cam->jpeg[1],cam->jpeg[2],cam->jpeg[3],cam->jpeg[4],cam->jpeg[5],cam->jpeg[6],cam->jpeg[7],cam->jpeg[8]);
  
  if (cam->skip) cam->skip--;
  else
    if ((n=jpeg_decode(cam->jpeg,cam->framebuf+framenum*DV3300_MAX_FRAME_SIZE,320,240,24,&cam->decdata))!=0)
      info("error decoding jpeg frame : %d",n);
  head=0;
  
  /* switch to BGR ? */
  for (i=0;i<320*240*3;i+=3)
  {
    swap=cam->framebuf[i];
    cam->framebuf[i]=cam->framebuf[i+2];
    cam->framebuf[i+2]=swap;
  }
    
  DBG("pic : %d %d %d %d %d %d %d %d %d",cam->framebuf[0],cam->framebuf[1],cam->framebuf[2],cam->framebuf[3],cam->framebuf[4],cam->framebuf[5],cam->framebuf[6],cam->framebuf[7],cam->framebuf[8]);

  /* brightness ... */
  DBG("brigthness=%d", cam->brightness);
  pos=cam->framebuf;
  for (i=0;i<320*240*3;i++,pos++)
  {
    temp=*pos+cam->brightness*2-256;
    if (temp<0) temp=0;
    if (temp>0xff) temp=0xff;
    *pos=(char)temp;
  }
  
  /* contrast ... */
  DBG("contrast=%d", cam->contrast);
  pos=cam->framebuf;
  for (i=0;i<320*240*3;i++,pos++)
  {
    temp=(*pos-128)*cam->contrast/128+128;
    if (temp<0) temp=0;
    if (temp>0xff) temp=0xff;
    *pos=(char)temp;
  }
  
  cam->nb++;
  
  temp=cam->nb;
  pos=text+13;
  while (temp)
  {
    *pos=temp%10+'0';
    temp=temp/10;
    pos--;
  }
  
  if (osd) osd_text(cam->framebuf,2,2,text);

  return;
}


static int dv3300_ioctl(struct video_device *dev,unsigned int ioctlnr,void *arg)
{
	int retval=0;
	struct dv3300 *cam = dev->priv;

	if (!cam)
		return -ENODEV;

	/* make this _really_ smp-safe */
/*	if (down_interruptible(&cam->busy_lock))
		return -EINTR;*/


	switch (ioctlnr)
  {

		/* query capabilites */
    case VIDIOCGCAP:
		{
			struct video_capability b;

			DBG("VIDIOCGCAP\n");
			strcpy(b.name, DRIVER_DESC " webcam");
			b.type = VID_TYPE_CAPTURE;
			b.channels = 1;
			b.audios = 0;
			b.maxwidth = 320;	/* VIDEOSIZE_CIF */
			b.maxheight = 240;
			b.minwidth = 320;	/* VIDEOSIZE_48_48 */
			b.minheight = 240;

			if (copy_to_user(arg, &b, sizeof (b)))
				retval = -EFAULT;

			break;
		}

		/* get/set video source - we are a camera and nothing else */
    case VIDIOCGCHAN:
		{
			struct video_channel v;

			DBG("VIDIOCGCHAN\n");
			if (copy_from_user(&v, arg, sizeof (v))) {
				retval = -EFAULT;
				break;
			}
			if (v.channel != 0) {
				retval = -EINVAL;
				break;
			}

			v.channel = 0;
			strcpy(v.name, "webcam");
			v.tuners = 0;
			v.flags = 0;
			v.type = VIDEO_TYPE_CAMERA;
			v.norm = 0;

			if (copy_to_user(arg, &v, sizeof (v)))
				retval = -EFAULT;
			break;
		}

    case VIDIOCSCHAN:
		{
			int v;

			if (copy_from_user(&v, arg, sizeof (v)))
				retval = -EFAULT;
			DBG("VIDIOCSCHAN %d\n", v);

			if (retval == 0 && v != 0)
				retval = -EINVAL;

			break;
		}

		/* image properties */
    case VIDIOCGPICT:
		{
			struct video_picture vp;
			DBG("VIDIOCGPICT\n");
			memset(&vp, 0, sizeof (struct video_picture));
			vp.brightness = cam->brightness << 8;
			vp.contrast = cam->contrast << 8;
			vp.depth = 24;
			vp.palette = VIDEO_PALETTE_RGB24;
			if (copy_to_user
			    (arg, &vp, sizeof (struct video_picture)))
				retval = -EFAULT;
			break;
		}

    case VIDIOCSPICT:
		{
			struct video_picture vp;
			
			if(copy_from_user(&vp, (struct video_picture *) arg,
				sizeof(struct video_picture)))
				retval = -EFAULT;

			else
			{
				DBG("VIDIOCSPICT depth = %d, pal = %d\n", vp.depth,
				    vp.palette);

				cam->brightness = vp.brightness >> 8;
        cam->contrast = vp.contrast >> 8;

				if (vp.depth != 24
				    || vp.palette != VIDEO_PALETTE_RGB24)
					retval = -EINVAL;
			}

			break;
		}

		/* get/set capture window */
    case VIDIOCGWIN:
		{
			struct video_window vw;
			vw.x = 0;
			vw.y = 0;
			vw.width = 320;
			vw.height = 240;
			vw.chromakey = 0;
			vw.flags = 0;
			vw.clips = NULL;
			vw.clipcount = 0;

			DBG("VIDIOCGWIN\n");

			if (copy_to_user
			    ((void *) arg, (void *) &vw, sizeof (vw)))
				retval = -EFAULT;

			// I'm not sure what the deal with a capture window is, it is very poorly described
			// in the doc.  So I won't support it now.
			break;
		}

    case VIDIOCSWIN:
		{

			struct video_window *vw = (struct video_window *) arg;
			DBG("VIDIOCSWIN %d x %d\n", vw->width, vw->height);

			if ( vw->width != 320 || vw->height != 240 )
				retval = -EFAULT;
			
			break;
		}

		/* mmap interface */
    case VIDIOCGMBUF:
		{
			struct video_mbuf vm;
			int i;

			DBG("VIDIOCGMBUF\n");
			memset(&vm, 0, sizeof (vm));
			vm.size =
			    DV3300_MAX_FRAME_SIZE * DV3300_FRAMES;
			vm.frames = DV3300_FRAMES;
			for (i = 0; i < DV3300_FRAMES; i++)
				vm.offsets[i] = DV3300_MAX_FRAME_SIZE * i;

			if (copy_to_user
			    ((void *) arg, (void *) &vm, sizeof (vm)))
				retval = -EFAULT;

			break;
		}

    case VIDIOCMCAPTURE:
		{
			struct video_mmap vm;
			// int video_size;

			if (copy_from_user
			    ((void *) &vm, (void *) arg, sizeof (vm))) {
				retval = -EFAULT;
				break;
			}

			DBG("VIDIOCMCAPTURE frame=%d, height=%d, width=%d, format=%d.\n",vm.frame,vm.width,vm.height,vm.format);

			if ( vm.frame >= DV3300_FRAMES || vm.format != VIDEO_PALETTE_RGB24 )
				retval = -EINVAL;

			// in theory right here we'd start the image capturing
			// (fill in a bulk urb and submit it asynchronously)
			//
			// Instead we're going to do a total hack job for now and
			// retrieve the frame in VIDIOCSYNC

			break;
		}

    case VIDIOCSYNC:
		{
			int frame;

			if (copy_from_user((void *) &frame, arg, sizeof (int))) {
				retval = -EFAULT;
				break;
			}
			DBG("VIDIOCSYNC: %d\n", frame);

			read_frame(cam, frame);
      
			break;
		}

      /* pointless to implement overlay with this camera */
    case VIDIOCCAPTURE:
    case VIDIOCGFBUF:
    case VIDIOCSFBUF:
    case VIDIOCKEY:
      retval = -EINVAL;
      break;
  
      /* tuner interface - we have none */
    case VIDIOCGTUNER:
    case VIDIOCSTUNER:
    case VIDIOCGFREQ:
    case VIDIOCSFREQ:
      retval = -EINVAL;
      break;
  
      /* audio interface - we have none */
    case VIDIOCGAUDIO:
    case VIDIOCSAUDIO:
      retval = -EINVAL;
      break;
    default:
      retval = -ENOIOCTLCMD;
      break;
	}

	//up(&cam->busy_lock);
	return retval;
}


/* v4l driver */
static struct video_device dv3300_template=
{
	owner:      THIS_MODULE,
	name:       DRIVER_DESC,
	type:       VID_TYPE_CAPTURE,
	hardware:   VID_HARDWARE_DV3300,
	open:       dv3300_open,
	close:      dv3300_close,
	read:       dv3300_read,
	write:      dv3300_write,
	ioctl:      dv3300_ioctl,
	mmap:       dv3300_mmap,
	initialize: dv3300_video_init,
	minor:-1
};


static void * dv3300_probe(struct usb_device *udev,unsigned int ifnum,const struct usb_device_id *id)
{
	struct dv3300 *cam=NULL;
	/*
  const struct usb_interface_descriptor *interface;
	const struct usb_endpoint_descriptor *endpoint;
  */

	if ((udev->descriptor.idVendor!=USB_DV3300_VENDOR_ID) || (udev->descriptor.idProduct!=USB_DV3300_PRODUCT_ID))
		return NULL;
  
  info(DRIVER_DESC " webcam plugged");

  if ((cam=kmalloc(sizeof(struct dv3300),GFP_KERNEL))==NULL)
  {
    info("cam: out of memory !");
    return NULL;
  }
  memset(cam,0x00,sizeof(struct dv3300));
  
  memcpy(&cam->vdev,&dv3300_template,sizeof(dv3300_template));
  cam->vdev.priv=cam;
  
  cam->udev=udev;

  if ((cam->buffer=kmalloc(BUFFER_SIZE,GFP_KERNEL))==NULL)
  {
    info("cam->buffer: out of memory !");
    return NULL;
  }
  
  if ((cam->jpeg=kmalloc(100000,GFP_KERNEL))==NULL)
  {
    info("cam->jpeg: out of memory !");
    return NULL;
  }
  
  cam->skip=2;
  cam->nb=0;
  cam->brightness=128;
  cam->contrast=128;
  
	/*
  cam->bulkEndpoint = bulkEndpoint;
  info("%d",bulkEndpoint);
  */

	if (video_register_device(&cam->vdev, VFL_TYPE_GRABBER, -1) == -1)
  {
		kfree(cam);
    info("video_register_device failed");
		return NULL;
	}
  
  info(DRIVER_DESC " controlling video device %d",cam->vdev.minor);
  return cam;
}


static void dv3300_disconnect(struct usb_device *udev,void *ptr)
{
  struct dv3300 *cam=ptr;
    
  info(DRIVER_DESC " webcam unplugged");
  video_unregister_device(&cam->vdev);
  kfree(cam->buffer);
  kfree(cam->jpeg);
  if (cam->framebuf) usbvideo_rvfree(cam->framebuf);
  kfree(cam);
}


static struct usb_driver dv3300_driver=
{
	name:       "dv3300",
	probe:      dv3300_probe,
	disconnect:	dv3300_disconnect,
  { NULL, NULL }
};


static int __init dv3300_init(void)
{
	if (usb_register(&dv3300_driver)<0)
    return -1;
  
  info(DRIVER_DESC " module loaded");
  return 0;
}


static void __exit dv3300_exit(void)
{
  info(DRIVER_DESC " module unloaded");
  usb_deregister(&dv3300_driver);
}


module_init(dv3300_init);
module_exit(dv3300_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
